"""
charsets
~~~~~~~~

Data and functions for working with character sets.
"""
from collections.abc import Iterator
from dataclasses import dataclass
from sys import byteorder

from charex import util


# Data classes.
@dataclass
class CodecDetails:
    """Information for working with the specific codec.

    :param size: (Optional.) The number of bytes used to address
        characters with the given character set.
    :param endian: (Optional.) The byte order used by the codec.
    """
    size: int = 1
    endian: str = byteorder
    description: str = ''


# Encoding schemes.
codecs = {
    'ascii': CodecDetails(
        description=(
            'RFC20 The ASCII format for Network Interchange.'
        )
    ),
    'big5': CodecDetails(
        description=(
            'The Big5 encoding method for traditional Chinese characters '
            'developed by the Institute for Information Industry of Taiwan '
            'in 1984.'
        )
    ),
    'big5hkscs': CodecDetails(
        description=(
            'Hong Kong Supplementary Character Set to the Big5 traditional '
            'Chinese character set.'
        )
    ),
    'cp037': CodecDetails(
        description=(
            'EBCDIC code page 37, USA/Canada Country Extended Code Page.'
        )
    ),
    'cp273': CodecDetails(
        description=(
            'EBCDIC code page 273, Germany/Austria.'
        )
    ),
    'cp424': CodecDetails(
        description=(
            'EBCDIC code page 424, Israel with supprt for Hebrew.'
        )
    ),
    'cp437': CodecDetails(
        description=(
            'Code page 424, default character set for the IBM PC.'
        )
    ),
    'cp500': CodecDetails(
        description=(
            'EBCDIC code page 500, full support of the Latin-1 character set.'
        )
    ),
    'cp720': CodecDetails(
        description=(
            'Code page 720, Arabic support for DOS.'
        )
    ),
    'cp737': CodecDetails(
        description=(
            'Code page 720, Greek support for DOS.'
        )
    ),
    'cp775': CodecDetails(
        description=(
            'Code page 775, Baltic language support for DOS.'
        )
    ),
    'cp850': CodecDetails(
        description=(
            'Code page 850, Western European language support for DOS.'
        )
    ),
    'cp852': CodecDetails(
        description=(
            'Code page 852, Central European language support for DOS.'
        )
    ),
    'cp855': CodecDetails(
        description=(
            'Code page 855, Cyrillic support for DOS.'
        )
    ),
    'cp856': CodecDetails(
        description=(
            'Code page 856, Hebrew language support for DOS.'
        )
    ),
    'cp857': CodecDetails(
        description=(
            'Code page 857, Turkish language support for DOS.'
        )
    ),
    'cp858': CodecDetails(
        description=(
            'Code page 858, Western European language support for DOS, '
            'modifying code page 850 by adding the Euro symbol.'
        )
    ),
    'cp860': CodecDetails(
        description=(
            'Code page 860, Portugese language support for DOS.'
        )
    ),
    'cp861': CodecDetails(
        description=(
            'Code page 861, Icelandic language support for DOS.'
        )
    ),
    'cp862': CodecDetails(
        description=(
            'Code page 862, Hebrew language support for DOS.'
        )
    ),
    'cp863': CodecDetails(
        description=(
            'Code page 863, Canadian French language support for DOS.'
        )
    ),
    'cp864': CodecDetails(
        description=(
            'Code page 864, Hebrew language support for DOS.'
        )
    ),
    'cp865': CodecDetails(
        description=(
            'Code page 865, Arabic language support for DOS.'
        )
    ),
    'cp866': CodecDetails(
        description=(
            'Code page 866, Nordic language support for DOS.'
        )
    ),
    'cp869': CodecDetails(
        description=(
            'Code page 869, Greek language support for DOS.'
        )
    ),
    'cp874': CodecDetails(
        description=(
            'Code page 874, Thai language support for DOS.'
        )
    ),
    'cp875': CodecDetails(
        description=(
            'EBCDIC code page 875, Greek.'
        )
    ),
    'cp932': CodecDetails(
        description=(
            'Code page 932, Japanese language support for Windows.'
        )
    ),
    'cp949': CodecDetails(
        description=(
            'Code page 949, Korean language support by IBM.'
        )
    ),
    'cp950': CodecDetails(
        description=(
            'Code page 932, Traditional Chinese language support for Windows.'
        )
    ),
    'cp1006': CodecDetails(
        description=(
            'Code page 1006, Urdu language support for AIX.'
        )
    ),
    'cp1026': CodecDetails(
        description=(
            'EBCDIC code page 1026, Turkish.'
        )
    ),
    'cp1125': CodecDetails(
        description=(
            'IBM code page 1125, Ukraine.'
        )
    ),
    'cp1140': CodecDetails(
        description=(
            'EBCDIC code page 1140, USA/Canada with Euro character.'
        )
    ),
    'cp1250': CodecDetails(
        description=(
            'Code page 1250, Central European language support for Windows.'
        )
    ),
    'cp1251': CodecDetails(
        description=(
            'Code page 1251, Cyrillic support for Windows.'
        )
    ),
    'cp1252': CodecDetails(
        description=(
            'Code page 1252, Latin-1 character set for Windows.'
        )
    ),
    'cp1253': CodecDetails(
        description=(
            'Code page 1253, Greek support for Windows.'
        )
    ),
    'cp1254': CodecDetails(
        description=(
            'Code page 1254, Turkish support for Windows.'
        )
    ),
    'cp1255': CodecDetails(
        description=(
            'Code page 1255, Hebrew support for Windows.'
        )
    ),
    'cp1256': CodecDetails(
        description=(
            'Code page 1256, Arabic support for Windows.'
        )
    ),
    'cp1257': CodecDetails(
        description=(
            'Code page 1257, Baltic language support for Windows.'
        )
    ),
    'cp1258': CodecDetails(
        description=(
            'Code page 1258, Vietamese support for Windows.'
        )
    ),
    'euc_jp': CodecDetails(
        description=(
            'Extended Unix Code Japanese.'
        )
    ),
    'euc_jis_2004': CodecDetails(
        description=(
            'Extended Unix Code Japanese Industrial Standard 2004.'
        )
    ),
    'euc_jisx0213': CodecDetails(
        description=(
            'Extended Unix Code Japanese Industrial Standard X 213.'
        )
    ),
    'euc_kr': CodecDetails(
        description=(
            'Extended Unix Code Korean.'
        )
    ),
    'gb2312': CodecDetails(
        description=(
            'Extended Unix Code Simplified Chinese.'
        )
    ),
    'gbk': CodecDetails(
        description=(
            'Extended Unix Code Simplified Chinese extended to include '
            'all unified CJK characters.'
        )
    ),
    'gb18030': CodecDetails(
        description=(
            'Chinese National Standard GB 18030-2005: Information '
            'Technology—Chinese coded character set.'
        )
    ),
    'hz': CodecDetails(
        description=(
            'Extended Unix Code Simplified Chinese for email.'
        )
    ),
    'iso2022_jp': CodecDetails(
        description=(
            'ISO 2022 standard for Japanese.'
        )
    ),
    'iso2022_jp_1': CodecDetails(
        description=(
            'ISO 2022 standard for Japanese, extension 1.'
        )
    ),
    'iso2022_jp_2': CodecDetails(
        description=(
            'ISO 2022 standard for Japanese, extension 2.'
        )
    ),
    'iso2022_jp_2004': CodecDetails(
        description=(
            'ISO 2022 standard for Japanese, extension 2004.'
        )
    ),
    'iso2022_jp_3': CodecDetails(
        description=(
            'ISO 2022 standard for Japanese, extension 3.'
        )
    ),
    'iso2022_jp_ext': CodecDetails(
        description=(
            'ISO 2022 standard for Japanese, extension.'
        )
    ),
    'iso2022_kr': CodecDetails(
        description=(
            'RFC1557 Korean Character Encoding for Internet Messages.'
        )
    ),
    'latin_1': CodecDetails(
        description=(
            'ISO-8859-1, Latin alphabet number 1 for western Europe.'
        )
    ),
    'iso8859_2': CodecDetails(
        description=(
            'ISO-8859-2, Latin alphabet number 2 for central Europe.'
        )
    ),
    'iso8859_3': CodecDetails(
        description=(
            'ISO-8859-3, Latin alphabet number 3 for southern Europe.'
        )
    ),
    'iso8859_4': CodecDetails(
        description=(
            'ISO-8859-4, Latin alphabet number 4 for northern Europe.'
        )
    ),
    'iso8859_5': CodecDetails(
        description=(
            'ISO-8859-5, Latin/Cyrillic alphabet.'
        )
    ),
    'iso8859_6': CodecDetails(
        description=(
            'ISO-8859-6, Latin/Arabic alphabet.'
        )
    ),
    'iso8859_7': CodecDetails(
        description=(
            'ISO-8859-7, Latin/Greek alphabet.'
        )
    ),
    'iso8859_8': CodecDetails(
        description=(
            'ISO-8859-8, Latin/Herbrew alphabet.'
        )
    ),
    'iso8859_9': CodecDetails(
        description=(
            'ISO-8859-9, Latin alphabet number 5 for Turkish.'
        )
    ),
    'iso8859_10': CodecDetails(
        description=(
            'ISO-8859-10, Latin alphabet number 6 for Nordic languages.'
        )
    ),
    'iso8859_11': CodecDetails(
        description=(
            'ISO-8859-11, Latin/Thai alphabet.'
        )
    ),
    'iso8859_13': CodecDetails(
        description=(
            'ISO-8859-13, Latin alphabet number 7 for Baltic Rim languages.'
        )
    ),
    'iso8859_14': CodecDetails(
        description=(
            'ISO-8859-14, Latin alphabet number 8 for Celtic languages.'
        )
    ),
    'iso8859_15': CodecDetails(
        description=(
            'ISO-8859-15, Latin alphabet number 9 for Western European '
            'languages, including the Euro symbol.'
        )
    ),
    'iso8859_16': CodecDetails(
        description=(
            'ISO-8859-16, Latin alphabet number 10 for south-eastern Europe.'
        )
    ),
    'johab': CodecDetails(
        description=(
            'KS X 1001 alternative character set for South Korean Hangul '
            'and Hanja.'
        )
    ),
    'koi8_r': CodecDetails(
        description=(
            'Kod Obmena Informatsiey, 8 bit, for Russian and Bulgarian.'
        )
    ),
    'koi8_t': CodecDetails(
        description=(
            'Kod Obmena Informatsiey, 8 bit, for Tajik Cyrillic.'
        )
    ),
    'koi8_u': CodecDetails(
        description=(
            'RFC2319 Ukrainian Character Set KOI8-U.'
        )
    ),
    'kz1048': CodecDetails(
        description=(
            'Windows-1251 variant for Kazakh.'
        )
    ),
    'mac_cyrillic': CodecDetails(
        description=(
            'Mac OS Cyrillic.'
        )
    ),
    'mac_greek': CodecDetails(
        description=(
            'Mac OS Greek.'
        )
    ),
    'mac_iceland': CodecDetails(
        description=(
            'Mac OS Icelandic.'
        )
    ),
    'mac_latin2': CodecDetails(
        description=(
            'Mac OS Central European, Microsoft code page 10029.'
        )
    ),
    'mac_roman': CodecDetails(
        description=(
            'Mac OS Western Europe.'
        )
    ),
    'mac_turkish': CodecDetails(
        description=(
            'Mac OS Turkish.'
        )
    ),
    'ptcp154': CodecDetails(
        description=(
            'Cyrillic-Asian.'
        )
    ),
    'shift_jis': CodecDetails(
        2,
        description=(
            'Japanese Industrial Standard with shifted first bytes.'
        )
    ),
    'shift_jis_2004': CodecDetails(
        2,
        description=(
            'Superset of Japanese Industrial Standard with shifted '
            'first bytes.'
        )
    ),
    'shift_jisx0213': CodecDetails(
        2,
        description=(
            'Superset of Japanese Industrial Standard with shifted '
            'first bytes.'
        )
    ),
    'utf_32': CodecDetails(
        4,
        description=(
            '32-bit Unicode Transformation format.'
        )
    ),
    'utf_32_be': CodecDetails(
        4,
        'big',
        description=(
            '32-bit Unicode Transformation format, big endian.'
        )
    ),
    'utf_32_le': CodecDetails(
        4,
        'little',
        description=(
            '32-bit Unicode Transformation format, little endian.'
        )
    ),
    'utf_16': CodecDetails(
        2,
        description=(
            '16-bit Unicode Transformation format.'
        )
    ),
    'utf_16_be': CodecDetails(
        2,
        'big',
        description=(
            '16-bit Unicode Transformation format, big endian.'
        )
    ),
    'utf_16_le': CodecDetails(
        2,
        'little',
        description=(
            '16-bit Unicode Transformation format, little endian.'
        )
    ),
    'utf_7': CodecDetails(
        description=(
            '7-bit Unicode Transformation format.'
        )
    ),
    'utf_8': CodecDetails(
        description=(
            '8-bit Unicode Transformation format.'
        )
    ),
    'utf_8_sig': CodecDetails(
        description=(
            '8-bit Unicode Transformation format, treating the BOM '
            'as metadata.'
        )
    ),
}


# Functions.
def get_codecs() -> tuple[str, ...]:
    """Return the keys of the registered codecs.

    :return: The keys of the codecs as a :class:`tuple`.
    :rtype: tuple
    """
    return tuple(codec for codec in codecs)


def get_codec_description(codec: str) -> str:
    """Provide the description for the given codec."""
    info = codecs[codec]
    return info.description


def multiencode(
    value: bytes | int | str,
    codecs_: Iterator[str]
) -> dict[str, bytes]:
    """Provide the address for the given character for each of the
    given character sets.

    :param value: The character to encode.
    :param codecs_: The codecs to encode to.
    :return: The encoded value for each character set as a :class:`dict`.
    :rtype: dict
    """
    value = util.to_char(value)
    results = {}
    for codec in codecs_:
        try:
            results[codec] = value.encode(codec)
        except UnicodeEncodeError:
            results[codec] = b''
    return results


def multidecode(
    value: int | str | bytes,
    codecs_: Iterator[str]
) -> dict[str, str]:
    """Provide the code point for the given address for each of the
    given character sets.

    :param value: The address to decode. This can be passed as either
        an :class:`int`, :class:`str`, or :class:`bytes`.
    :param codec_: The codecs to decode to.
    :return: The decoded value for each character set as a :class:`dict`.
    :rtype: dict
    """
    # Coerce the given value into bytes.
    value = util.to_bytes(value)

    # Decode the value into the character sets.
    results = {}
    for codec in codecs_:
        b = value

        # Pad for 2 or 4 byte codecs.
        while len(b) < codecs[codec].size:
            if codecs[codec].endian == 'little':
                b = b + b'\x00'
            else:
                b = b'\x00' + b

        # Decode.
        try:
            results[codec] = b.decode(codec)
        except UnicodeDecodeError:
            results[codec] = ''
    return results



"""
test_shell
~~~~~~~~~~

Unit tests for :mod:`charex.shell`.
"""
from charex import escape as esc
from charex import normal as nl
from charex import shell as sh


# Tests for cd.
def test_cd(capsys):
    """Invoked with a hex string, `cd` returns the character for
    that address in each registered character set. A hex string is
    declared by starting the string with "0x".
    """
    with open('tests/data/charset_mode_41.txt') as fh:
        exp = fh.read()
    cmd = 'cd 0x41'
    shell_test(exp, cmd, capsys)


def test_cd_binary(capsys):
    """Invoked with a binary string, `cd` returns the character for
    that address in each registered character set. A binary string is
    declared by starting the string with "0b".
    """
    with open('tests/data/charset_mode_41.txt') as fh:
        exp = fh.read()
    cmd = 'cd 0b01000001'
    shell_test(exp, cmd, capsys)


def test_cd_string(capsys):
    """Invoked with a character, `cd` returns the character for
    that character's UTF-8 address in each registered character set.
    """
    with open('tests/data/charset_mode_41.txt') as fh:
        exp = fh.read()
    cmd = 'cd A'
    shell_test(exp, cmd, capsys)


# Tests for ce.
def test_ce(capsys):
    """Invoked with a character, `ce` returns the address for
    that character in each registered character set.
    """
    with open('tests/data/charset_mode_r.txt') as fh:
        exp = fh.read()
    cmd = 'ce A'
    shell_test(exp, cmd, capsys)


# Tests for cl.
def test_cl(capsys):
    """Invoked, `cl` returns the list of registered character sets."""
    with open('tests/data/charsetlist.txt') as fh:
        exp = fh.read()
    cmd = 'cl'
    shell_test(exp, cmd, capsys)


def test_cl_description(capsys):
    """Invoked with "-d", `cl` returns the list of registered character
    sets with descriptions.
    """
    with open('tests/data/charsetlist_d.txt') as fh:
        exp = fh.read()
    cmd = 'cl -d'
    shell_test(exp, cmd, capsys)


# Tests for ct.
def test_ct(capsys):
    """Invoked with a normalization form and a base string, ct mode
    should print the number of denormalizations using the given form to
    stdout.
    """
    exp = '120,270,240\n\n'
    cmd = 'ct nfkd <script>'
    shell_test(exp, cmd, capsys)


def test_ct_maxdepth(capsys):
    """Invoked with "-m" and an integer, ct mode limit the number of
    denormalizations per character to the given integer and print the
    number of denormalizations using the given form to stdout.
    """
    exp = '256\n\n'
    cmd = 'ct nfkd <script> -m 2'
    shell_test(exp, cmd, capsys)


# Test dn mode.
def test_dn(capsys):
    """Invoked with a normalization form and a base string, dn mode
    should print the denormalizations for the base string to stdout.
    """
    # Expected result.
    exp = (
        '\ufe64\ufe63\ufe65\n'
        '\ufe64\ufe63\uff1e\n'
        '\ufe64\uff0d\ufe65\n'
        '\ufe64\uff0d\uff1e\n'
        '\uff1c\ufe63\ufe65\n'
        '\uff1c\ufe63\uff1e\n'
        '\uff1c\uff0d\ufe65\n'
        '\uff1c\uff0d\uff1e\n'
        '\n'
    )
    cmd = (
        'dn '
        'nfkd '
        '<->'
    )
    shell_test(exp, cmd, capsys)


def test_dn_number(capsys):
    """Invoked with -n and an integer, dn mode should return no
    more than that number of results.
    """
    exp = (
        '\ufe64\ufe63\ufe65\n'
        '\n'
    )
    cmd = (
        'dn '
        'nfkd '
        '<-> '
        '-m 1'
    )
    shell_test(exp, cmd, capsys)


def test_dn_random(capsys):
    """Called with -r, dn mode should return a randomly
    denormalize the string.
    """
    exp = (
        '﹤－﹥\n'
        '\n'
    )
    cmd = (
        'dn '
        'nfkd '
        '<-> '
        '-r '
        '-s spam'
    )
    shell_test(exp, cmd, capsys)


# Test dt mode.
def test_dt(capsys):
    """Invoked with a character, details mode should print the details
    for the character.
    """
    with open('tests/data/details_mode_A.txt') as fh:
        exp = fh.read()
    cmd = (
        'dt '
        'A'
    )
    shell_test(exp, cmd, capsys)


# Test el mode.
def test_el(capsys):
    """When invoked, el mode returns a list of the registered
    escape schemes.
    """
    exp = '\n'.join(scheme for scheme in esc.schemes) + '\n\n'
    cmd = 'el'
    shell_test(exp, cmd, capsys)


def test_el_description(capsys):
    """Invoked with "-d", `el` returns the list of registered character
    sets with descriptions.
    """
    with open('tests/data/el_d.txt') as fh:
        exp = fh.read()
    cmd = 'el -d'
    shell_test(exp, cmd, capsys)


# Test es mode.
def test_es(capsys):
    """Invoked with a scheme and a base string, escape mode should
    escape the string using the given scheme and print the escaped
    string.
    """
    exp = '%41\n\n'
    cmd = (
        'es '
        'url '
        'A'
    )
    shell_test(exp, cmd, capsys)


# Test fl mode.
def test_fl(capsys):
    """When invoked, fl mode returns a list of the registered
    normalization forms.
    """
    exp = '\n'.join(form for form in nl.forms) + '\n\n'
    cmd = (
        'fl'
    )
    shell_test(exp, cmd, capsys)


def test_fl_description(capsys):
    """When invoked with -d, fl mode should return a list of
    registered normalization forms and a brief description of each
    one.
    """
    with open('tests/data/fl_d.txt') as fh:
        exp = fh.read()
    cmd = (
        'fl '
        '-d'
    )
    shell_test(exp, cmd, capsys)


# Test nl mode.
def test_nl(capsys):
    """When invoked with a normalization form and a base string,
    nl mode returns the normalization of the base string using the
    given form.
    """
    exp = 'A\n\n'
    cmd = (
        'nl '
        'nfkc '
        '\u24b6'
    )
    shell_test(exp, cmd, capsys)


# Utility functions.
def shell_test(exp, cmd, capsys):
    """Test shell invocation."""
    shell = sh.Shell()
    shell.onecmd(cmd)
    captured = capsys.readouterr()
    assert captured.out == exp



"""
util
~~~~

Utility functions for :mod:`charex`.
"""
from importlib.resources import files
from math import log
import unicodedata as ucd


# Constants.
RESOURCES = {
    # Command help.
    'help_xt': 'help_xt.txt',

    # Denormalization lookups.
    'rev_casefold': 'rev_casefold.json',
    'rev_nfc': 'rev_nfc.json',
    'rev_nfkc': 'rev_nfkc.json',
    'rev_nfd': 'rev_nfd.json',
    'rev_nfkd': 'rev_nfkd.json',

    # Unicode data.
    'propvals': 'PropertyValueAliases.txt',
    'unicodedata': 'UnicodeData.txt',

    # HTML data.
    'entities': 'entities.json',

    # HTML examples.
    'result': 'result.html',
    'quote': 'quote.html',
}


# Exceptions.
class NotCharacterError(ValueError):
    """The value cannot be used as a character."""


# Functions
def bin2bytes(value: str, endian: str = 'big') -> bytes:
    """Convert a binary string into :class:`bytes`.

    :param value: A :class:`str` containing the representation of
        a binary number.
    :param endian: (Optional.) An indicator for the endianness of the
        binary number. Valid values are: big, little. It defaults to
        big.
    :return: The binary number as :class:`bytes`.
    :rtype: bytes
    """
    value = pad_byte(value, endian, base=2)

    parts = []
    while value:
        parts.append(value[:8])
        value = value[8:]
    nums = [int(s, 2) for s in parts]
    octets = [n.to_bytes((n.bit_length() + 7) // 8) for n in nums]
    return b''.join(octets)


def get_description_from_docstring(obj: object) -> str:
    """Get the first paragraph of the docstring from the given object.

    :param obj: An object with a docstring.
    :return: The first paragraph of the object's docstring as a :class:`str`.
    :rtype: str
    """
    doc = obj.__doc__
    if doc:
        paragraphs = doc.split('\n\n')
        descr = paragraphs[0]
        lines = descr.split('\n')
        lines = [line.lstrip() for line in lines]
        return ' '.join(lines)
    return ''


def hex2bytes(value: str, endian: str = 'big') -> bytes:
    """Convert a hex string into :class:`bytes`.

    :param value: A :class:`str` containing the representation of
        a hexadecimal number.
    :param endian: (Optional.) An indicator for the endianness of the
        hexadecimal number. Valid values are: big, little. It defaults
        to big.
    :return: The hexadecimal number as :class:`bytes`.
    :rtype: bytes
    """
    # Since a byte is two characters, pad strings that have an
    # odd length.
    value = pad_byte(value, endian)

    # Convert the string to bytes.
    parts = []
    while value:
        parts.append(value[:2])
        value = value[2:]
    nums = [int(s, 16) for s in parts]
    octets = [n.to_bytes((n.bit_length() + 7) // 8) for n in nums]
    return b''.join(octets)


def neutralize_control_characters(value: str) -> str:
    """Transform control characters in a string into the Unicode
    symbol for those characters.

    :param value: The :class:`str` to neutralize.
    :return: The neutralized :class:`str`.
    :rtype: str
    """
    def neutralize(char: str) -> str:
        try:
            if ucd.category(char) == 'Cc':
                num = ord(char)
                new = chr(num + 0x2400)
                return new
            return char
        except TypeError as ex:
            print(char)
            raise ex

    return ''.join(neutralize(char) for char in value)


def pad_byte(value: str, endian: str = 'big', base: int = 16) -> str:
    """Add a zeros to pad strings shorter than the needed bytelen.

    :param value: A :class:`str` containing the representation of
        a number.
    :param endian: (Optional.) An indicator for the endianness of the
        number. Valid values are: big, little. It defaults to big.
    :param base: (Optional.) The base of the number. It defaults to
        hexadecimal (16).
    :return: The number padded with leading zeros to be a full byte
        as a :class:`str`.
    :rtype: str
    """
    # Determine the number of digits needed in a byte.
    bytelen = int(log(256, base))

    # Pad the number.
    if gap := len(value) % bytelen:
        zeros = '0' * (bytelen - gap)
        if endian == 'big':
            return zeros + value
        return value[:-1 * gap] + zeros + value[-1 * gap:]
    return value


def read_resource(key: str, codec: str = 'utf_8') -> tuple[str, ...]:
    """Read the data from a resource file within the package.

    :param key: The key for the file in the RESOURCES constant.
    :return: The contents of the file as a :class:`tuple`.
    :rtype: tuple
    """
    pkg = files('charex.data')
    data_file = pkg / RESOURCES[key]
    fh = data_file.open(encoding=codec)
    lines = fh.readlines()
    fh.close()
    return tuple(lines)


def to_bytes(value: bytes | int | str, endian: str = 'big') -> bytes:
    """Transform the given value to :class:`bytes`.

    :param value: The value to transform.
    :param endian: (Optional.) An indicator for the endianness of the
        a number string. Valid values are: big, little. It defaults to
        big.
    :return: A :class:`bytes`.
    :rtype: bytes
    """
    if isinstance(value, str) and value.startswith('0b'):
        value = bin2bytes(value[2:], endian)
    elif isinstance(value, str) and value.startswith('0x'):
        value = hex2bytes(value[2:], endian)
    elif isinstance(value, str) and value.startswith('U+'):
        n = int(value[2:], 16)
        char = chr(n)
        value = char.encode('utf8')
    elif isinstance(value, str):
        value = value.encode('utf8')
    elif isinstance(value, int):
        value = value.to_bytes((value.bit_length() + 7) // 8)
    return value


def to_char(value: bytes | int | str) -> str:
    """Transform the given value to a one character :class:`str`.

    :param value: The value to transform.
    :return: A :class:`str` of length one.
    :rtype: str
    """
    prefixes = {
        '0b': 2,
        '0d': 10,
        '0o': 8,
        '0x': 16,
        'U+': 16,
    }

    if isinstance(value, bytes):
        value = value.decode('utf8')
    elif isinstance(value, int):
        value = chr(value)
    elif value[:2] in prefixes:
        n = int(value[2:], prefixes[value[:2]])
        value = chr(n)
    elif not isinstance(value, str) or len(value) != 1:
        msg = 'Value cannot be made a str with length one.'
        raise NotCharacterError(msg)
    return value



"""
test_normal
~~~~~~~~~~~

Unit testing for :mod:`charex.normal`.
"""
from charex import normal as nl


# Tests for form_casefold().
def test_form_casefold():
    """Given a base string, :class:`charex.normal.form_casefold` should
    return the base string normalized with the casefold form.
    """
    exp = 'a'
    base = 'A'
    assert nl.form_casefold(base) == exp


# Tests for form_nfc().
def test_form_nfc():
    """Given a base string, :class:`charex.normal.form_nfc` should
    return the base string normalized with the NFC form.
    """
    exp = 'é'
    base = '\u0065\u0301'
    assert nl.form_nfc(base) == exp


def test_form_nfd():
    """Given a base string, :class:`charex.normal.form_nfd` should
    return the base string normalized with the NFD form.
    """
    exp = '\u0065\u0301'
    base = 'é'
    assert nl.form_nfd(base) == exp


# Tests for form_nfkc().
def test_form_nfkd():
    """Given a base string, :class:`charex.normal.form_nfkc` should
    return the base string normalized with the NFKC form.
    """
    exp = 'A'
    base = '\u24b6'
    assert nl.form_nfkc(base) == exp


# Tests for form_nfkd().
def test_form_nfkd():
    """Given a base string, :class:`charex.normal.form_nfkd` should
    return the base string normalized with the NFKD form.
    """
    exp = 'A'
    base = '\u24b6'
    assert nl.form_nfkd(base) == exp


# Tests for get_description().
def test_get_description():
    """Given the key for an escape scheme,
    :func:`charex.normal.get_description` should
    return the first paragraph of the docstring of
    that scheme.
    """
    # Expected value.
    exp = 'Eggs bacon.'

    # Test set up.
    @nl.reg_form('__test_get_description')
    def form_spam(char, codec):
        """Eggs bacon.

        Ham and toast.
        """
        return char

    # Run test and determine result.
    assert nl.get_description('__test_get_description') == exp

    # Test clean up.
    del nl.forms['__test_get_description']


# Tests for normalize().
def test_normalize():
    """Given a normalization schem and base string, return the
    base string normalized using the normalization scheme as a
    :class:`str`.
    """
    exp = 'A'
    scheme = 'nfkc'
    base = '\u24b6'
    assert nl.normalize(scheme, base) == exp


# Tests for build_denormalization_map().
def test_build_denormalization_map():
    """When given a denormalization function,
    :func:`charex.normal.build_denormalization_map` will return a map
    of every character that normalizes into a character as a
    JSON string.
    """
    with open('charex/data/rev_nfc.json') as fh:
        exp = fh.read()[:-2]
    form = nl.form_nfc
    act = nl.build_denormalization_map(form)
    alines = act.split('\n')
    elines = exp.split('\n')
    for i, lines in enumerate(zip(alines, elines)):
        a, e = lines
        assert (i, a) == (i, e)
    assert act == exp


# Tests for find_max_decomposition().
def test_find_max_decomposition():
    """When called, :func:`charex.normal.find_max_decomposition` finds
    the character that decomposes into the largest string, then returns
    that character and the length of the string it decomposes into as
    a tuple.
    """
    exp = ('ᾂ', 4)
    assert nl.find_max_decomposition() == exp



"""
denormal
~~~~~~~~

Functions for reversing normalization of string.
"""
from collections.abc import Generator, Sequence
from math import prod
from random import choice, seed

from charex.charex import Character


# Functions.
def count_denormalizations(
    base: str,
    form: str,
    maxdepth: int | None = None
) -> int:
    """Determine the number of denormalizations that exist for the string.

    :param base: The :class:`str` to denormalize.
    :param form: The Unicode normalization form to denormalize from.
        Valid values are: casefold, nfc, nfd, nfkc, nfkd.
    :param maxdepth: (Optional.) How many individual characters to use
        when denormalizing the base. This is used to limit the total
        number of denormalizations of the overall base.
    :return: The number of denormalizations as an :class:`int`.
    :rtype: int

    Usage
    -----
    To count the number of possible denormalizations for a given string
    and form::

        >>> base = '<->'
        >>> form = 'nfkc'
        >>> count_denormalizations(base, form)
        8

    """
    chars = (Character(c) for c in base)
    counts = []
    for char in chars:
        count = len(char.denormalize(form))
        if count == 0:
            count = 1
        if maxdepth and count > maxdepth:
            count = maxdepth
        counts.append(count)
    return int(prod(counts))


def denormalize(
    base: str,
    form: str,
    maxdepth: int = 0,
    maxresults: int | None = None,
    random: bool = False,
    seed_: bytes | int | str = ''
) -> tuple[str, ...]:
    """Denormalize a string.

    :param base: The :class:`str` to denormalize.
    :param form: The Unicode normalization form to denormalize from.
        Valid values are: casefold, nfc, nfd, nfkc, nfkd.
    :param maxdepth: (Optional.) How many denormalizations per character
        in the base string to use when denormalizing the base. This is
        used to limit the total number of denormalizations of the overall
        base. If `maxdepth` is zero, the number of denormalizations to
        use per character is not limited.
    :param maxresults: (Optional.) The maximum number of results to
        return. Default behavior varies based on the `random` parameter.
        If `random` is `False`, default is to return all possible
        denormalizattions. Otherwise, the default is to return one.
    :param random: (Optional.) Whether to pick randomly from the
        possible denormalization results. Defaults to false.
    :param seed: (Optional.) A seed value for the random number generator.
        Defaults to not seeding the generator.
    :return: The denormalizations as a :class:`tuple`.
    :rtype: tuple

    Usage
    -----
    To denormalize a given string with the given form::

        >>> base = '<>'
        >>> form = 'nfkc'
        >>> denormalize(base, form)
        ('﹤﹥', '﹤＞', '＜﹥', '＜＞')

    The `maxdepth` parameter can be used to limit the number of
    denormalizations per character in the `base` string. This is
    useful when you want just a few denormalizations of a string
    with a very large number of denormalizations::

        >>> base = 'hi'
        >>> form = 'nfkc'
        >>> maxdepth = 2
        >>> denormalize(base, form, maxdepth)
        ('ʰᵢ', 'ʰⁱ', 'ₕᵢ', 'ₕⁱ')

    """
    if random:
        return random_denormalize(base, form, maxresults, seed_)

    # Get the denormalized forms of the first character.
    char = Character(base[0])
    dechars = list(char.denormalize(form))

    # If there are no denormalized forms, then it is the denormalized form.
    if not dechars:
        dechars = [char.value,]

    # Limit the number of permutations by limiting the number of
    # denormalized forms we are looking at.
    if maxdepth and len(dechars) > maxdepth:
        dechars = dechars[:maxdepth]

    # If there are more characters left, use recursion to get the
    # permutations for those characters, then get the permutations
    # with this character.
    if base[1:]:
        results = []
        tails = denormalize(base[1:], form, maxdepth, maxresults)
        for dechar in dechars:
            for tail in tails:
                results.append(dechar + tail)

    # If there are no characters left, then the permutations are just
    # the denormalized forms of the character.
    else:
        results = dechars

    # Truncate to the maximum results and return.
    if maxresults:
        results = results[:maxresults]
    return tuple(results)


def gen_denormalize(
    base: str,
    form: str,
    maxdepth: int = 0
) -> Generator[str, None, None]:
    """Denormalize a string, yielding the results as they are
    generated.

    :param base: The :class:`str` to denormalize.
    :param form: The Unicode normalization form to denormalize from.
        Valid values are: casefold, nfc, nfd, nfkc, nfkd.
    :param maxdepth: (Optional.) How many denormalizations per character
        in the base string to use when denormalizing the base. This is
        used to limit the total number of denormalizations of the overall
        base. If `maxdepth` is zero, the number of denormalizations to
        use per character is not limited.
    :return: A :class:`collections.abc.Generator` that yields the
        denormalization results.
    :rtype: collections.abc.Generator

    Usage
    -----
    To generate denormalizations for a given string with a given form::

        >>> base = '<>'
        >>> form = 'nfkc'
        >>> dngen = gen_denormalize(base, form)
        >>> [result for result in dngen]
        ['﹤﹥', '﹤＞', '＜﹥', '＜＞']

    The `maxdepth` parameter can be used to limit the number of
    denormalizations per character in the `base` string. This is
    useful when you want just a few denormalizations of a string
    with a very large number of denormalizations::

        >>> base = 'hi'
        >>> form = 'nfkc'
        >>> maxdepth = 2
        >>> dngen = gen_denormalize(base, form, maxdepth)
        >>> [result for result in dngen]
        ['ʰᵢ', 'ʰⁱ', 'ₕᵢ', 'ₕⁱ']

    """
    c, rest = base[0], base[1:]
    char = Character(c)
    dechars = char.denormalize(form)
    if not dechars:
        dechars = (char.value,)
    if maxdepth:
        dechars = dechars[:maxdepth]

    if rest:
        for dechar in dechars:
            for tail in gen_denormalize(rest, form, maxdepth):
                yield dechar + tail

    else:
        for dechar in dechars:
            yield dechar


def gen_random_denormalize(
    base: str,
    form: str,
    maxresults: int = 1,
    seed_: bytes | int | str = ''
) -> Generator[str, None, None]:
    """Randomly denormalize a string, yielding the results as they
    are generated. This is useful when returning all results for
    a denormalization is unreasonably large, as can easily happen
    when denormalizing strings containing Latin letters.

    :param base: The :class:`str` to denormalize.
    :param form: The Unicode normalization for to denormalize from.
        Valid values are: NFC, NFD, NFKC, NFKD.
    :param maxresults: (Optional.) The maximum number of results to
        return. The default is to return one.
    :param seed: (Optional.) A seed value for the random number generator.
        Defaults to not seeding the generator.
    :return: A :class:`collections.abc.Generator` that yields the random
        denormalization results.
    :rtype: collections.abc.Generator
    """
    chars = [denormalize(char, form) for char in base]
    if seed_:
        seed(seed_)
    for _ in range(maxresults):
        result = ''.join(choice(char) for char in chars)
        yield result


def random_denormalize(
    base: str,
    form: str,
    maxresults: int | None = None,
    seed_: bytes | int | str = ''
) -> tuple[str, ...]:
    """Randomly denormalize a string. This is useful when returning
    all results for a denormalization is unreasonably large, as can
    easily happen when denormalizing strings containing Latin letters.

    :param base: The :class:`str` to denormalize.
    :param form: The Unicode normalization for to denormalize from.
        Valid values are: NFC, NFD, NFKC, NFKD.
    :param maxresults: (Optional.) The maximum number of results to
        return. Default behavior varies based on the `random` parameter.
        If `random` is `False`, default is to return all possible
        denormalizattions. Otherwise, the default is to return one.
    :param seed: (Optional.) A seed value for the random number generator.
        Defaults to not seeding the generator.
    :return: The denormalizations as a :class:`tuple`.
    :rtype: tuple
    """
    # Ensure at least one result is returned.
    if not maxresults:
        maxresults = 1

    # Seeding the RNG allows for repeatability in testing.
    if seed_:
        seed(seed_)

    # Get the denormalized forms for all the characters in the string.
    chars = [
        denormalize(char, form)
        for char in base
    ]

    # Randomly pick from the possible denormalizations for each character
    # when creating the denormalized strings, then return the results.
    results = []
    for _ in range(maxresults):
        result = ''.join(choice(char) for char in chars)
        results.append(result)
    return tuple(results)



"""
charex
~~~~~~

Tools for exploring unicode characters and other character sets.
"""
from collections.abc import Sequence
from dataclasses import dataclass
from json import loads
import unicodedata as ucd

from charex import util
from charex.escape import schemes


# Data classes.
@dataclass
class UnicodeDatum:
    """A record from the UnicodeData.txt file for Unicode 14.0.0.

    :param code_point: The address for the character in Unicode.
    :param name: The name for the code point.
    :param category: The type of code point, such as "control" or
        "lower case letter."
    :param canonical_combining_class: The combining class of the code point,
        largely used for CJK languages.
    :param bidi_class: Unknown.
    :param decomposition_type: Whether and how the character can be
        decomposed.
    :param decimal: If the character is a decimal digit, this is its
        numeric value.
    :param digit: If the character is a digit, this is its numeric
        value.
    :param numeric: If the character is a number, this is its numeric
        value.
    :param bidi_mirrored: Unknown.
    :param unicode_1_name: The name of the character used in Unicode
        version 1. This is mainly needed to give names to control
        characters.
    :param iso_comment: Unknown.
    :param simple_uppercase_mapping: The code point for the upper case
        version of the code point.
    :param simple_lowercase_mapping: The code point for the lower case
        version of the code point.
    :param simple titlecase_mapping: The code point for the title case
        version of the code point.
    """
    code_point: str
    name: str
    general_category: str
    canonical_combining_class: str
    bidi_class: str
    decomposition_type: str
    decimal: str
    digit: str
    numeric: str
    bidi_mirrored: str
    unicode_1_name: str
    iso_comment: str
    simple_uppercase_mapping: str
    simple_lowercase_mapping: str
    simple_titlecase_mapping: str


# Caches.
propvals_cache: dict[str, dict[str, str]] = {}
unicodedata_cache: dict[str, UnicodeDatum] = {}


# Utility functions.
def expand_property_value(alias: str, proptype: str) -> str:
    """Translate the short name of a Unicode property value into the
    long name for that property.

    :param alias: The short name to translate.
    :param proptype: The type of property.
    :return: The long name of the property as a :class:`str`.
    :rtype: str
    """
    # Look it up in the cache, to avoid having to reload the file
    # multiple times.
    try:
        by_alias = propvals_cache[proptype]

    # If it's not in the cache, then we have to load the data from
    # file.
    except KeyError:
        lines = util.read_resource('propvals')
        by_alias = parse_property_values(lines, proptype)
        propvals_cache[proptype] = by_alias

    # Return the expanded alias.
    return by_alias[alias]


# Data parsing functions.
def parse_property_values(
    lines: Sequence[str],
    proptype: str
) -> dict[str, str]:
    """Parse the contents of the property values file and return the
    translation map for the given property type.

    :param lines: The contents of the property values file.
    :param proptype: The type of properties to extract from the file.
    :return: The entries for the given property type as a :class:`dict`.
    :rtype: dict
    """
    lines = [line for line in lines if line.startswith(proptype)]
    by_alias = {}
    for line in lines:
        line = line.split('#', 1)[0]
        fields = line.split(';')
        key = fields[1].strip()
        value = fields[2].strip()
        value = value.replace('_', ' ')
        by_alias[key] = value
    return by_alias


def parse_unicode_data(lines: Sequence[str]) -> dict[str, UnicodeDatum]:
    """Parse the Unicode data file.

    :param lines: The contents of the Unicode data file.
    :return: The Unicode data as a :class:`dict`.
    :rtype: dict
    """
    if not unicodedata_cache:
        for line in lines:
            fields = line.split(';')
            datum = UnicodeDatum(*fields)
            unicodedata_cache['U+' + datum.code_point] = datum
    return unicodedata_cache


# Classes.
class Character:
    """One or more code points representing a character.

    :param value: The character to gather data for. See below for the
        formats the value can be passed in.
    :return: None.
    :rtype: NoneType

    Character Formats
    =================
    The value can be passed in a couple of different formats:

    *   *Single-Character string*: The value is a :class:`str` of
        length one.
    *   *Hex string*: The value is a hexadecimal number between 0x00
        and 0x10FFFF passed as a :class:`str` prefixed with "0x".
    *   *Address string*: The value is a hexadecimal number between
        0x00 and 0x10FFFF passed as a :class:`str` prefixed with "U+".
    """
    def __init__(self, value: str) -> None:
        value = util.to_char(value)
        self.__value = value
        self._rev_normal_cache: dict[str, tuple[str, ...]] = {}

    def __repr__(self) -> str:
        return f'{self.code_point} ({self.name})'

    @property
    def category(self) -> str:
        """The Unicode general category for the character."""
        alias = ucd.category(self.value)
        return expand_property_value(alias, 'gc')

    @property
    def code_point(self) -> str:
        """The address for the character in the Unicode database."""
        x = ord(self.value)
        return f'U+{x:04x}'.upper()

    @property
    def decimal(self) -> int | None:
        """The decimal value of the character."""
        return ucd.decimal(self.value, None)

    @property
    def decomposition(self) -> str:
        """The Unicode defined decompositions of the character."""
        return ucd.decomposition(self.value)

    @property
    def digit(self) -> int | None:
        """The numerical value of the character as a digit."""
        return ucd.digit(self.value, None)

    @property
    def name(self) -> str:
        """The Unicode name for the character."""
        try:
            name = ucd.name(self.value)

        # Control characters don't have assigned names in Unicode
        # 14.0.0. So, we have to look up the Unicode 1 names for
        # them, which are in the 14.0.0 UnicodeData.txt file.
        except ValueError:
            cat = ucd.category(self.value)

            # Control characters.
            if cat == 'Cc':
                if not unicodedata_cache:
                    lines = util.read_resource('unicodedata')
                    data = parse_unicode_data(lines)
                else:
                    data = unicodedata_cache
                point = self.code_point
                name = f'<{data[point].unicode_1_name}>'

            # Private use characters.
            elif cat == 'Co':
                name = 'PRIVATE USE CHARACTER'

            # Fall back if there are more code points without names.
            else:
                name = '?? UNKNOWN ??'

        return name

    @property
    def numeric(self) -> float | int | None:
        """The Unicode defined numeric value for the character."""
        return ucd.numeric(self.value, None)

    @property
    def value(self) -> str:
        """The code point as a string."""
        return self.__value

    def denormalize(self, form: str) -> tuple[str, ...]:
        """Return the characters that normalize to the character using
        the given form.

        :param form: The normalization form to check against.
        :return: The denormalization results in a :class:`tuple`.
        :rtype: tuple
        """
        source = f'rev_{form}'
        if source not in self._rev_normal_cache:
            lkp = Lookup(source)
            self._rev_normal_cache[source] = lkp.query(self.value)
        return self._rev_normal_cache[source]

    def escape(self, scheme: str, codec: str = 'utf8') -> str:
        """The escaped version of the character.

        :param scheme: The escape scheme to use.
        :param codec: The codec to use when escaping to a hexadecimal
            string.
        :return: A :class:`str` with the escaped character.
        :rtype: str
        """
        scheme = scheme.casefold()
        fn = schemes[scheme]
        return fn(self.value, codec)

    def encode(self, codec: str) -> str:
        """The hexadecimal value for the character in the given
        character set.

        :param codec: The codec to use when encoding to a hexadecimal
            string.
        :return: A :class:`str` with the encoded character.
        :rtype: str
        """
        b = self.value.encode(codec)
        hexes = [f'{x:02x}'.upper() for x in b]
        return ''.join(x for x in hexes)

    def is_normal(self, form: str) -> bool:
        """Is the character normalized to the given form?

        :param form: The normalization form to check against.
        :return: A :class:`bool` indicating whether the character is
            normalized.
        :rtype: bool
        """
        return ucd.is_normalized(form, self.value)

    def normalize(self, form: str) -> str:
        """Normalize the character using the given form.

        :param form: The normalization form to check against.
        :return: The normalization result as a :class:`str`.
        :rtype: str
        """
        return ucd.normalize(form, self.value)

    def summarize(self) -> str:
        """Return a summary of the character's information.

        :return: The character information as a :class:`str`.
        :rtype: str
        """
        value = util.neutralize_control_characters(self.value)
        return f'{value} {self!r}'


class Lookup:
    """A data lookup.

    :param source: The key for the data source in the RESOURCES
        dictionary.
    """
    def __init__(self, source: str) -> None:
        self.__source = source
        lines = util.read_resource(source)
        json = '\n'.join(lines)
        data = loads(json)
        self.__data = {k: tuple(data[k]) for k in data}

    @property
    def data(self) -> dict[str, tuple[str, ...]]:
        """The data loaded from the data source."""
        return self.__data

    @property
    def source(self) -> str:
        """The key of the loaded data in the RESOURCES dictionary."""
        return self.__source

    def query(self, key: str) -> tuple[str, ...]:
        """Return the value for the given string from the loaded data.

        :param key: The key to lookup in the data.
        :return: The data returned in a :class:`tuple`.
        :rtype: tuple
        """
        try:
            answer = self.data[key]
        except KeyError:
            answer = tuple()
        return answer



"""
cmds
~~~~

Core logic for the different commands/modes of :mod:`charex`.
"""
from collections.abc import Callable, Generator, Sequence
from itertools import zip_longest
from textwrap import wrap

from charex import charex as ch
from charex import charsets as cset
from charex import denormal as dnm
from charex import escape as esc
from charex import normal as nml
from charex import util


# Command functions.
def cd(address: str) -> Generator[str, None, None]:
    """Decode the given address in all codecs.

    :param address: The address to decode in the codecs.
    :return: Yields the result for each codec as a :class:`str`.
    :rtype: str
    """
    # Get the data.
    codecs = cset.get_codecs()
    results = cset.multidecode(address, (codec for codec in codecs))

    # Write the output.
    width = max(len(codec) for codec in codecs)
    for key in results:
        c = results[key]
        details = ''
        if len(c) < 1:
            details = '*** no character ***'
        elif len(c) > 1:
            details = '*** multiple characters ***'
        else:
            char = ch.Character(c)
            details = f'{char.code_point} {char.name}'
        c = util.neutralize_control_characters(c)
        yield f'{key:>{width}}: {c} {details}'


def ce(base: str) -> Generator[str, None, None]:
    """Encode the given character in all codecs.

    :param address: The character to encode in the codecs.
    :return: Yields the result for each codec as a :class:`str`.
    :rtype: str
    """
    # Get the data.
    codecs = cset.get_codecs()
    results = cset.multiencode(base, (codec for codec in codecs))

    # Write the output.
    width = max(len(codec) for codec in codecs)
    for key in results:
        if b := results[key]:
            c = ' '.join(f'{n:>02x}'.upper() for n in b)
            yield f'{key:>{width}}: {c}'


def cl(show_descr: bool = False) -> Generator[str, None, None]:
    """List registered character sets.

    :param show_descr: (Optional.) Whether to show the descriptions
        for the character sets.
    :return: Yields each codec as a :class:`str`.
    :rtype: str
    """
    codecs = cset.get_codecs()
    for line in write_list(codecs, cset.get_codec_description, show_descr):
        yield line


def ct(base: str, form: str, maxdepth: int, number: int = 0) -> str:
    """Count denormalization results.

    :param base: The base normalized string.
    :param form: The Unicode normalization form for the denormalization.
    :param maxdepth: Maximum number of reverse normalizations to use
        for each character.
    :param number: (Optional.) The number of denormalizations to return.
    :return: The number of denormalizations as a :class:`str`
    :rtype: str
    """
    if number:
        return f'{number:,}'
    count = dnm.count_denormalizations(base, form, maxdepth)
    return f'{count:,}'


def dn(
    base: str,
    form: str,
    maxdepth: int,
    random: bool,
    seed: bytes | int | str
) -> Generator[str, None, None]:
    """Perform denormalizations.

    :param base: The base normalized string.
    :param form: The normalization form for the denormalization.
    :param maxdepth: If not random, sets the maximum number of
        denormalizations to use for each character. If random, sets
        the number of random denormalizations to return.
    :param random: Randomize the denormalization.
    :param seed: Seed the randomized denormalization.
    :return: Yields each codec as a :class:`str`.
    :rtype: str
    """
    if not random:
        for result in dnm.gen_denormalize(base, form, maxdepth):
            yield result

    else:
        if not maxdepth:
            maxdepth = 1
        for result in dnm.gen_random_denormalize(
            base,
            form,
            maxdepth,
            seed
        ):
            yield result


def dt(c: str) -> Generator[str, None, None]:
    """Display details for a code point.

    :param address: The character to encode in the codecs.
    :return: Yields the result for each codec as a :class:`str`.
    :rtype: str
    """
    def rev_normalize(char: ch.Character, form: str) -> str:
        points = char.denormalize(form)
        values = []
        for point in points:
            if len(point) == 1:
                c = ch.Character(point)
                values.append(c.summarize())
            elif len(point) > 1:
                values.append(f'{point} *** multiple characters ***')
                for item in point:
                    char = ch.Character(item)
                    values.append('  ' + char.summarize())
        if not values:
            return ''
        return ('\n' + ' ' * 22).join(v for v in values)

    # Gather the details for display.
    char = ch.Character(c)
    details = (
        ('Display', char.value),
        ('Name', char.name),
        ('Code Point', char.code_point),
        ('Category', char.category),
        ('UTF-8', char.encode('utf8')),
        ('UTF-16', char.encode('utf_16_be')),
        ('UTF-32', char.encode('utf_32_be')),
        ('Decomposition', char.decomposition),
        ('C encoded', char.escape('c')),
        ('URL encoded', char.escape('url')),
        ('HTML encoded', char.escape('html')),
        ('Reverse Cfold', rev_normalize(char, 'casefold')),
        ('Reverse NFC', rev_normalize(char, 'nfc')),
        ('Reverse NFD', rev_normalize(char, 'nfd')),
        ('Reverse NFKC', rev_normalize(char, 'nfkc')),
        ('Reverse NFKD', rev_normalize(char, 'nfkd')),
    )

    # Display the details.
    width = 20
    for detail in details:
        label, value = detail
        if value:
            yield f'{label:>{width}}: {value}'


def el(show_descr: bool = False) -> Generator[str, None, None]:
    """List registered escape schemes.

    :param show_descr: (Optional.) Whether to show the descriptions
        for the character sets.
    :return: Yields each codec as a :class:`str`.
    :rtype: str
    """
    schemes = esc.get_schemes()
    for line in write_list(schemes, esc.get_description, show_descr):
        yield line


def es(base: str, scheme: str, codec: str) -> str:
    """Escape a string using the given scheme.

    :param base: The string to escape.
    :param scheme: The key in the `schemes` :class:`dict` to use for
        the escaping.
    :param codec: The character set codec to use when escaping the
        characters.
    :return: The escaped :class:`str`.
    :rtype: str
    """
    return esc.escape(base, scheme, codec)


def fl(show_descr: bool = False) -> Generator[str, None, None]:
    """List registered normalization forms.

    :param show_descr: (Optional.) Whether to show the descriptions
        for the character sets.
    :return: Yields each form as a :class:`str`.
    :rtype: str
    """
    forms = nml.get_forms()
    for line in write_list(forms, nml.get_description, show_descr):
        yield line


def nl(form: str, base: str, expand: bool = False) -> str:
    """Perform normalizations.

    :param form: The key of a registered normalization form.
    :param base: The string to normalize.
    :param expand: (Optional.) Whether to provide a summary of each
        character in the normalization.
    :return: The normalized :class:`str`.
    :rtype: str
    """
    result = nml.normalize(form, base)
    out = result
    if expand:
        out += '\n'
        for item in result:
            char = ch.Character(item)
            indent = '  '
            if 'mark' in char.category.casefold():
                indent += ' '
            out += f'  {char.summarize()}\n'
    return out


# Utility functions.
def make_description_row(name: str, namewidth: int, descr: str) -> str:
    """Create a two column row with a name and description.

    :param name: The content for the first column.
    :param namewidth: The width of the first column.
    :param descr: The content for the second column.
    :return: The row as a :class:`str`.
    :rtype: str
    """
    name_lines = wrap(name, namewidth)
    descr_lines = wrap(descr, 77 - namewidth)
    lines = (
        f'{n:<{namewidth}}  {d}'
        for n, d in zip_longest(name_lines, descr_lines, fillvalue='')
    )
    return '\n'.join(lines)


def write_list(
    items: Sequence[str],
    get_descr: Callable[[str], str],
    show_descr: bool
) -> Generator[str, None, None]:
    """Output the given list.

    :param items: The items to list.
    :param get_descr: The function used to look up the discription
        of each item.
    :param show_descr: Whether include the descriptions of each item
        in the output.
    :return: Yields each codec as a :class:`str`.
    :rtype: str
    """
    width = max(len(item) for item in items)
    for item in items:
        if show_descr:
            descr = get_descr(item)
            row = make_description_row(item, width, descr)
            yield row
        else:
            yield item



"""
gui
~~~

A graphical user interface for :mod:`charex`.
"""
import tkinter as tk
from tkinter import ttk

from charex import charex as ch
from charex import cmds
from charex import charsets as cset
from charex import denormal as dn
from charex import escape as esc
from charex import normal as nl
from charex import util
from charex import shell as sh


# Constants.
ALL = (tk.N, tk.E, tk.W, tk.S)
SIDES = (tk.W, tk.E)
ENDS = (tk.N, tk.S)


# Application classes.
class Application:
    """The GUI for :mod:`charex`."""
    # Initialization.
    def __init__(self, root):
        # Configure the main window.
        self.root = root
        root.title('charex')
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)

        # Create the tab navigation and tabs.
        book = ttk.Notebook(root, padding='1 1 1 1')
        self.book = book
        book.grid(column=0, row=0, sticky=ALL)
        self.tabs = {}
        self.wake_focus = {}
        names = ('cd', 'ce', 'cl', 'ct', 'dn', 'dt', 'el', 'es', 'fl', 'nl')
        for i, name in enumerate(names):
            frame = ttk.Frame(book, padding='3 3 12 12')
            book.add(frame, text=name)
            init = getattr(self, f'init_{name}')
            num = i + 1
            if num == 1:
                num = ''
            init(frame, num)
            self.tabs[i] = getattr(self, name)

        # Bind event handlers.
        root.bind('<Return>', self.handle_return)
        root.bind('<<NotebookTabChanged>>', self.handle_notebook_tab_changed)

    def init_cd(self, frame, num=None):
        """Initialize the "cd" tab.

        :param frame: The frame for the "cd" notebook tab.
        :return: None.
        :rtype: NoneType
        """
        self.cd_address = tk.StringVar()
        self.cd_result = self.make_results(frame)

        self.config_simple_grid(frame)
        address_entry = self.make_entry(frame, self.cd_address)
        cd_button = self.make_button(frame, 'Decode', self.cd)
        self.pad_kids(frame)
        address_entry.focus_set()
        self.wake_focus[f'!frame{num}'] = address_entry

    def init_ce(self, frame, num=None):
        self.ce_char = tk.StringVar()
        self.ce_result = self.make_results(frame)

        self.config_simple_grid(frame)
        char_entry = self.make_entry(frame, self.ce_char)
        cd_button = self.make_button(frame, 'Encode', self.ce)
        self.pad_kids(frame)
        self.wake_focus[f'!frame{num}'] = char_entry

    def init_cl(self, frame, num=None):
        self.cl_result = self.make_results(frame)

        self.config_simple_grid(frame)
        cl_button = self.make_button(frame, 'List Character Sets', self.cl)
        self.pad_kids(frame)

    def init_ct(self, frame, num=None):
        self.ct_base = tk.StringVar()
        self.ct_form = tk.StringVar()
        self.ct_maxdepth = tk.StringVar()
        self.ct_maxdepth.set('0')
        self.ct_result = self.make_results(frame, row=5, colspan=4)

        self.config_five_params_grid(frame)

        char_entry = self.make_entry(
            frame,
            self.ct_base,
            colspan=5
        )

        form_label = ttk.Label(frame, text='Form:', justify=tk.RIGHT)
        form_label.grid(column=0, row=2, columnspan=1, sticky=SIDES)
        form_combo = ttk.Combobox(frame, textvariable=self.ct_form)
        form_combo['values'] = nl.get_forms()
        form_combo.state(['readonly'])
        form_combo.grid(column=1, row=2, columnspan=1, sticky=SIDES)

        maxdepth_label = ttk.Label(frame, text='Max Depth:', justify=tk.RIGHT)
        maxdepth_label.grid(column=2, row=2, columnspan=1, sticky=SIDES)
        maxdepth_entry = self.make_entry(
            frame,
            self.ct_maxdepth,
            width=40,
            col=3,
            row=2,
            colspan=2
        )

        cd_button = self.make_button(
            frame,
            'Count Denormalizations',
            self.ct,
            row=4,
            colspan=5
        )
        self.pad_kids(frame)
        self.wake_focus[f'!frame{num}'] = char_entry

    def init_dn(self, frame, num=None):
        self.dn_base = tk.StringVar()
        self.dn_form = tk.StringVar()
        self.dn_maxdepth = tk.StringVar()
        self.dn_maxdepth.set('0')
        self.dn_random = tk.BooleanVar(value=False)
        self.dn_seed = tk.StringVar()
        self.dn_result = self.make_results(frame, row=5, colspan=4)

        self.config_five_params_grid(frame)

        char_entry = self.make_entry(
            frame,
            self.dn_base,
            colspan=5
        )

        form_label = ttk.Label(frame, text='Form:', justify=tk.RIGHT)
        form_label.grid(column=0, row=2, columnspan=1, sticky=SIDES)
        form_combo = ttk.Combobox(frame, textvariable=self.dn_form)
        form_combo['values'] = nl.get_forms()
        form_combo.state(['readonly'])
        form_combo.grid(column=1, row=2, columnspan=1, sticky=SIDES)

        maxdepth_label = ttk.Label(frame, text='Max Depth:', justify=tk.RIGHT)
        maxdepth_label.grid(column=2, row=2, columnspan=1, sticky=SIDES)
        maxdepth_entry = self.make_entry(
            frame,
            self.dn_maxdepth,
            width=40,
            col=3,
            row=2,
            colspan=2
        )

        random_label = ttk.Label(frame, text='Random:', justify=tk.RIGHT)
        random_label.grid(column=0, row=3, columnspan=1, sticky=SIDES)
        random_check = ttk.Checkbutton(
            frame,
            variable=self.dn_random,
            onvalue='True',
            offvalue='False'
        )
        random_check.grid(column=1, row=3, columnspan=1, sticky=SIDES)

        seed_label = ttk.Label(frame, text='Seed:', justify=tk.RIGHT)
        seed_label.grid(column=2, row=3, columnspan=1, sticky=SIDES)
        seed_entry = self.make_entry(
            frame,
            self.dn_seed,
            width=40,
            col=3,
            row=3,
            colspan=2
        )

        cd_button = self.make_button(
            frame,
            'Denormalize',
            self.dn,
            row=4,
            colspan=5
        )
        self.pad_kids(frame)
        self.wake_focus[f'!frame{num}'] = char_entry

    def init_dt(self, frame, num=None):
        """Initialize the "dt" tab.

        :param frame: The frame for the "dt" notebook tab.
        :return: None.
        :rtype: NoneType
        """
        self.dt_char = tk.StringVar()
        self.dt_result = self.make_results(frame)

        self.config_simple_grid(frame)
        address_entry = self.make_entry(frame, self.dt_char)
        cd_button = self.make_button(frame, 'Character Details', self.dt)
        self.pad_kids(frame)
        self.wake_focus[f'!frame{num}'] = address_entry

    def init_el(self, frame, num=None):
        self.el_result = self.make_results(frame)

        self.config_simple_grid(frame)
        el_button = self.make_button(frame, 'List Escape Schemes', self.el)
        self.pad_kids(frame)

    def init_es(self, frame, num=None):
        self.es_base = tk.StringVar()
        self.es_scheme = tk.StringVar()
        self.es_result = self.make_results(frame, row=5, colspan=4)

        self.config_five_params_grid(frame)

        char_entry = self.make_entry(
            frame,
            self.es_base,
            colspan=5
        )

        scheme_label = ttk.Label(frame, text='Scheme:', justify=tk.RIGHT)
        scheme_label.grid(column=0, row=2, columnspan=1, sticky=SIDES)
        scheme_combo = ttk.Combobox(frame, textvariable=self.es_scheme)
        scheme_combo['values'] = esc.get_schemes()
        scheme_combo.state(['readonly'])
        scheme_combo.grid(column=1, row=2, columnspan=4, sticky=SIDES)

        es_button = self.make_button(
            frame,
            'Escape',
            self.es,
            row=4,
            colspan=5
        )
        self.pad_kids(frame)
        self.wake_focus[f'!frame{num}'] = char_entry

    def init_fl(self, frame, num=None):
        self.fl_result = self.make_results(frame)

        self.config_simple_grid(frame)
        el_button = self.make_button(
            frame,
            'List Normalization Forms',
            self.fl
        )
        self.pad_kids(frame)

    def init_nl(self, frame, num=None):
        self.nl_base = tk.StringVar()
        self.nl_form = tk.StringVar()
        self.nl_result = self.make_results(frame, row=5, colspan=4)

        self.config_five_params_grid(frame)

        char_entry = self.make_entry(
            frame,
            self.nl_base,
            colspan=5
        )

        form_label = ttk.Label(frame, text='Form:', justify=tk.RIGHT)
        form_label.grid(column=0, row=2, columnspan=1, sticky=SIDES)
        form_combo = ttk.Combobox(frame, textvariable=self.nl_form)
        form_combo['values'] = nl.get_forms()
        form_combo.state(['readonly'])
        form_combo.grid(column=1, row=2, columnspan=4, sticky=SIDES)

        cd_button = self.make_button(
            frame,
            'Normalize',
            self.nl,
            row=4,
            colspan=5
        )
        self.pad_kids(frame)
        self.wake_focus[f'!frame{num}'] = char_entry

    # Core commands.
    def cd(self, *args):
        try:
            self.cd_result.delete('0.0', 'end')
            address = self.cd_address.get()
            for line in cmds.cd(address):
                self.cd_result.insert('end', line + '\n')

        except ValueError:
            ...

    def ce(self, *args):
        try:
            self.ce_result.delete('0.0', 'end')
            base = self.ce_char.get()
            for line in cmds.ce(base):
                self.ce_result.insert('end', line + '\n')

        except ValueError:
            ...

    def cl(self, *args):
        self.cl_result.delete('0.0', 'end')
        for line in cmds.cl(True):
            self.cl_result.insert('end', line + '\n\n')

    def ct(self, *args):
        self.ct_result.delete('0.0', 'end')
        base = self.ct_base.get()
        form = self.ct_form.get()
        maxdepth = int(self.ct_maxdepth.get())
        line = cmds.ct(base, form, maxdepth)
        self.ct_result.insert('end', line + '\n\n')

    def dn(self, *args):
        self.dn_result.delete('0.0', 'end')
        base = self.dn_base.get()
        form = self.dn_form.get()
        maxdepth = int(self.dn_maxdepth.get())
        random = self.dn_random.get()
        seed_ = self.dn_seed.get()

        if not random:
            for line in dn.gen_denormalize(base, form, maxdepth):
                self.dn_result.insert('end', line + '\n')

        else:
            for line in dn.gen_random_denormalize(
                base,
                form,
                maxdepth,
                seed_
            ):
                self.dn_result.insert('end', line + '\n')

    def dt(self, *args):
        try:
            self.dt_result.delete('0.0', 'end')
            base = self.dt_char.get()
            for line in cmds.dt(base):
                self.dt_result.insert('end', line + '\n')

        except ValueError:
            ...

    def el(self, *args):
        self.el_result.delete('0.0', 'end')
        for line in cmds.el(True):
            self.el_result.insert('end', line + '\n\n')

    def es(self, *args):
        self.es_result.delete('0.0', 'end')
        base = self.es_base.get()
        scheme = self.es_scheme.get()
        line = cmds.es(base, scheme, 'utf8')
        self.es_result.insert('end', line)

    def fl(self, *args):
        self.fl_result.delete('0.0', 'end')
        for line in cmds.fl(True):
            self.fl_result.insert('end', line + '\n\n')

    def nl(self, *args):
        self.nl_result.delete('0.0', 'end')
        base = self.nl_base.get()
        form = self.nl_form.get()

        result = cmds.nl(form, base, True)
        self.nl_result.insert('end', result)

    # Event handlers.
    def handle_notebook_tab_changed(self, event):
        focus = self.root.focus_get()
        name = str(focus)
        frame = name.split('.')[-2]
        if frame in self.wake_focus:
            entry = self.wake_focus[frame]
            entry.focus_set()

    def handle_return(self, *args):
        tab_id = self.book.select()
        tab = self.book.index(tab_id)
        cmd = self.tabs[tab]
        cmd()

    # Grid configuration.
    def config_simple_grid(self, frame):
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=0)
        frame.rowconfigure(1, weight=0)
        frame.rowconfigure(2, weight=0)
        frame.rowconfigure(3, weight=1)

    def config_five_params_grid(self, frame):
        frame.columnconfigure(0, weight=0)
        frame.columnconfigure(1, weight=1)
        frame.columnconfigure(2, weight=0)
        frame.columnconfigure(3, weight=1)
        frame.columnconfigure(4, weight=0)
        frame.rowconfigure(1, weight=0)
        frame.rowconfigure(2, weight=0)
        frame.rowconfigure(3, weight=0)
        frame.rowconfigure(4, weight=0)
        frame.rowconfigure(5, weight=1)

    # Generic widget creation.
    def make_button(
        self,
        frame,
        name,
        cmd,
        col=0,
        row=2,
        colspan=2,
        rowspan=1,
        sticky=SIDES
    ):
        button = ttk.Button(frame, text=name, command=cmd)
        button.grid(
            column=col,
            row=row,
            columnspan=colspan,
            rowspan=rowspan,
            sticky=sticky
        )
        return button

    def make_entry(
        self,
        frame,
        value,
        width=80,
        col=0,
        row=1,
        colspan=2,
        rowspan=1,
        sticky=SIDES,
        justify=tk.RIGHT
    ):
        entry = ttk.Entry(
            frame,
            width=width,
            textvariable=value,
            justify=tk.RIGHT
        )
        entry.grid(
            column=col,
            row=row,
            columnspan=colspan,
            rowspan=rowspan,
            sticky=sticky
        )
        return entry

    def make_results(
        self,
        frame,
        row=3,
        colspan=1
    ):
        text = tk.Text(frame, width=80, height=24, wrap='word')
        ys = ttk.Scrollbar(
            frame,
            orient='vertical',
            command=text.yview
        )
        text['yscrollcommand'] = ys.set
        text.grid(column=0, row=row, columnspan=colspan, sticky=ALL)
        ys.grid(column=colspan, row=row, sticky=ENDS)
        return text

    # Generic frame configuration.
    def pad_kids(self, frame):
        for child in frame.winfo_children():
            child.grid_configure(padx=2, pady=4)


def main():
    root = tk.Tk()
    app = Application(root)
    root.mainloop()


if __name__ == '__main__':
    main()



"""
test_util
~~~~~~~~~

Unit test for :mod:`charex.util`.
"""
import pytest

from charex import util


# Test cases.
# Tests for to_bytes.
def test_to_bytes():
    """Given a :class:`str` containing a representation of a binary
    number, return that number as :class:`bytes`.
    """
    exp = b'\xe9'
    value = '0b11101001'
    act = util.to_bytes(value)
    assert act == exp


def test_to_bytes_bin_len_not_multiple_of_eight():
    """Given a :class:`str` containing a representation of a binary
    number, return that number as :class:`bytes`. If the length of the
    :class:`str` is not a multiple of eight, prepend enough zeros to
    make it a multiple of eight.
    """
    exp = b'\x29\xe9'
    value = '0b10100111101001'
    act = util.to_bytes(value)
    assert act == exp


def test_to_bytes_len_not_multiple_of_eight_little_endian():
    """Given a :class:`str` containing a representation of a binary
    number, return that number as :class:`bytes`. If the length of
    the :class:`str` is not a multiple of eight and the string is
    little endian, prepend enough zeros to the last byte to make it
    a multiple of eight.
    """
    exp = b'\xa7\x29'
    value = '0b10100111101001'
    act = util.to_bytes(value, endian='little')
    assert act == exp


def test_to_bytes_bytes():
    """Given a :class:`bytes`, :func:`charex.util.bytes` should return
    the given :class:`bytes`.
    """
    exp = b'\x00\x61'
    assert util.to_bytes(exp) == exp


def test_to_bytes_hex():
    """Given a :class:`str` containing a representation of a hexadecimal
    number, return that number as :class:`bytes`.
    """
    exp = b'\xbe\xef\xca\x5e'
    value = '0xbeefca5e'
    act = util.to_bytes(value)
    assert act == exp


def test_hex2bytes_hex_odd_length():
    """Given a :class:`str` containing a representation of a hexadecimal
    number, return that number as :class:`bytes`. If the :class:`str` has
    an odd number of characters, prepend a zero to the string before
    converting it to bytes.
    """
    exp = b'\x0e\xef\xca\x5e'
    value = '0xeefca5e'
    act = util.to_bytes(value)
    assert act == exp


def test_hex2bytes_hex_odd_length_little_endian():
    """Given a :class:`str` containing a representation of a hexadecimal
    number, return that number as :class:`bytes`. If the :class:`str` has
    an odd number of characters and endian is "little" add a zero before
    the last character of the string.
    """
    exp = b'\xbe\xef\xca\x0e'
    value = '0xbeefcae'
    act = util.to_bytes(value, endian='little')
    assert act == exp


def test_to_bytes_int():
    """Given a :class:`int`, :func:`charex.util.bytes` should that
    number as :class:`bytes`.
    """
    exp = b'\x61'
    value = 0x61
    assert util.to_bytes(value) == exp


def test_to_bytes_str():
    """Given a :class:`str`, return that :class:`str` as an UTF-8
    encoded :class:`bytes`.
    """
    exp = b'\xc3\xa9'
    value = 'é'
    act = util.to_bytes(value)
    assert act == exp


def test_to_bytes_unicode():
    """Given a :class:`str` containing a Unicode code point, return that
    code point as an UTF-8 encoded :class:`bytes`.
    """
    exp = b'\xe9'
    value = '0b11101001'
    act = util.to_bytes(value)
    assert act == exp


# Tests for to_char.
def to_char():
    """Given a :class:`str` with length one, return that :class:`str`."""
    exp = 'a'
    assert util.to_char(exp) == exp


def test_sto_char_binary():
    """Given an binary string, return a one character :class:`str` using
    the given binary string as the unicode code point.
    """
    exp = 'a'
    value = '0b01100001'
    assert util.to_char(value) == exp


def test_to_char_bytes():
    """Given :class:`bytes`, return a one character :class:`str` using
    the given bytes as a UTF-8 encoded value.
    """
    exp = '\U0001f600'
    value = b'\xf0\x9f\x98\x80'
    assert util.to_char(value) == exp


def test_sto_char_decimal():
    """Given an decimal string, return a one character :class:`str` using
    the given decimal string as the unicode code point.
    """
    exp = 'a'
    value = '0d97'
    assert util.to_char(value) == exp


def test_sto_char_hex():
    """Given a hex string, return a one character :class:`str` using
    the given hex string as the unicode code point.
    """
    exp = 'a'
    value = '0x61'
    assert util.to_char(value) == exp


def test_to_char_int():
    """Given an :class:`int`, return a one character :class:`str` using
    the given :class:`int` as the Unicode code point.
    """
    exp = 'a'
    value = 0x61
    assert util.to_char(value) == exp


def test_sto_char_octal():
    """Given an octal string, return a one character :class:`str` using
    the given octal string as the unicode code point.
    """
    exp = 'a'
    value = '0o141'
    assert util.to_char(value) == exp


def test_to_char_str_too_long():
    """Given a :class:`str` with more than one character that isn't a
    hex string or a code point, :func:`charex.util.to_char` should
    raise :class:`charex.util.NotCharacterError`.
    """
    exp = util.NotCharacterError
    value = 'spam'
    with pytest.raises(exp):
        util.to_char(value)


def test_to_char_unicode():
    """Given a Unicode code point string, return a one character
    :class:`str` of the unicode code point.
    """
    exp = 'a'
    value = 'U+0061'
    assert util.to_char(value) == exp



"""
test_denormal
~~~~~~~~~~~~~

Unit tests for :mod:`charex.denormal`.
"""
from charex import denormal as d


# denormalize tests.
def test_denormalize():
    """When given a string and a normalization form,
    :func:`charex.denormal.denormal` returns a tuple of
    strings that will normalize to the given string.
    """
    exp = (
        '\ufe64\ufe63\ufe65',
        '\ufe64\ufe63\uff1e',
        '\ufe64\uff0d\ufe65',
        '\ufe64\uff0d\uff1e',
        '\uff1c\ufe63\ufe65',
        '\uff1c\ufe63\uff1e',
        '\uff1c\uff0d\ufe65',
        '\uff1c\uff0d\uff1e',
    )
    act = d.denormalize('<->', 'nfkc')
    assert exp == act


def test_denormalize_maxcount():
    """When given a string, a normalization form, and a maximum number
    of results, :func:`charex.denormal.denormal` returns a tuple of
    strings that will normalize to the given string, limiting the number
    returned by the maximum number of results.
    """
    exp = (
        '\ufe64\ufe63\ufe65',
        '\ufe64\ufe63\uff1e',
        '\ufe64\uff0d\ufe65',
        '\ufe64\uff0d\uff1e',
    )
    act = d.denormalize('<->', 'nfkc', maxresults=4)
    assert exp == act


def test_denormalize_maxdepth():
    """When given a string, a normalization form, and a maximum depth,
    :func:`charex.denormal.denormal` returns a tuple of strings that
    will normalize to the given string, limiting the recursion by the
    maximum depth.
    """
    exp = (
        '\ufe64\ufe63\ufe65',
    )
    act = d.denormalize('<->', 'nfkc', 1)
    assert exp == act


def test_denormalize_random():
    """When given a string, a normalization form, a maximum number
    of results, a randomization flag, and a seed value,
    :func:`charex.denormal.denormal` returns a tuple of strings that
    will normalize to the given string, randomly selected from the
    possibilities.
    """
    exp = (
        '﹤－﹥',
        '＜－＞',
        '＜－＞',
        '﹤﹣＞',
    )
    act = d.denormalize(
        '<->',
        'nfkc',
        maxresults=4,
        random=True,
        seed_='spam'
    )
    assert exp == act


# count_denormalizations tests.
def test_count_denormalizations():
    """Given a string and a normalization form,
    :func:`charex.denormal.count_denormalizations`
    returns the number of strings that will normalize
    to the given string.
    """
    assert d.count_denormalizations('<->', 'nfkc') == 8


def test_count_denormalizations_maxdepth():
    """When given a string, a normalization form, and a maximum depth,
    :func:`charex.denormal.count_denormalizations` returns the number
    of strings that will normalize to the given string, limiting the
    recursion by the maximum depth.
    """
    assert d.count_denormalizations('<->', 'nfkc', 1) == 1



"""
http
~~~~

A web interface for mod:`charex`.
"""
from unicodedata import normalize

from flask import Flask, make_response, request, send_from_directory

from charex import charex


# Create the web application.
app = Flask(__name__)


# Utility functions.
def search_data(query: str) -> tuple[str, ...]:
    raise RuntimeError('Calamity!')


def shutdown_server():
    """Shutdown the server."""
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


def sanitize(value: str) -> str:
    """Sanitize input."""
    bad = ["'", '"', '<']
    for char in bad:
        if char in value:
            raise ValueError('Disallowed character used.')
    return value


# Responders.
@app.route('/badresult', methods=['GET',])
def badresult():
    """The vulnerable result response."""
    # Validate input.
    untrust = request.args['q']
    clean = sanitize(untrust)            # Sanitization avoids bad characters.

    # Perform the search.
    normal = normalize('NFKC', clean)    # Normalization can undo sanitization.
    try:
        result = search_data(normal)
        # Cutting lines to format results for display.
        return result

    # Display error if search errors.
    except RuntimeError:
        lines = charex.read_resource('result')
        doc = ''.join(lines)
        resp = make_response(doc.format(normal))    # Unsanitary input.
        return resp


@app.route('/goodresult', methods=['GET',])
def goodresult():
    """The less vulnerable result response."""
    # Validate input.
    untrust = request.args['q']
    normal = normalize('NFKC', untrust)  # Normalization first.
    clean = sanitize(normal)             # Sanitization avoids bad characters.

    # Perform the search.
    try:
        result = search_data(clean)
        # Cutting lines to format results for display.
        return result

    # Display error if search errors.
    except RuntimeError:
        lines = charex.read_resource('result')
        doc = ''.join(lines)
        resp = make_response(doc.format(clean))     # Sanitary input.
        return resp


@app.route('/quote', methods=['GET',])
def quote():
    """Demonstrate character set differences."""
    codec = 'cp1252'
    if 'charset' in request.args:
        codec = request.args['charset']
    lines = charex.read_resource('quote', codec=codec)
    doc = ''.join(line for line in lines)
    resp = make_response(doc)
    return resp


@app.route('/shutdown', methods=['GET',])
def shutdown():
    """Process request to shutdown the server."""
    shutdown_server()
    return 'Shutting down'


@app.route('/styles/<path:path>')
def styles(path):
    """Serve static files."""
    return send_from_directory('static/styles', path)


# Mainline.
if __name__ == '__main__':
    app.run(port=8080, debug=True)
